/*
 * Copyright (c) 2020 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package vcl

import "text/template"

const selfShardTmplSrc = `
import std;
import directors;

probe vk8s_probe_varnish {
	.request = "HEAD /vk8s_cluster_health HTTP/1.1"
	           "Host: vk8s_cluster"
	           "Connection: close";
{{- if .Probe.Timeout}}
	.timeout = {{.Probe.Timeout}};
{{- end}}
{{- if .Probe.Interval}}
	.interval = {{.Probe.Interval}};
{{- end}}
{{- if .Probe.Initial}}
	.initial = {{.Probe.Initial}};
{{- end}}
{{- if .Probe.Window}}
	.window = {{.Probe.Window}};
{{- end}}
{{- if .Probe.Threshold}}
	.threshold = {{.Probe.Threshold}};
{{- end}}
}

{{range $node := .Nodes -}}
backend {{$node.Name}} {
	.host = "{{(index $node.Addresses 0).IP}}";
	.port = "{{(index $node.Addresses 0).Port}}";
	.probe = vk8s_probe_varnish;
}

{{end -}}

acl vk8s_cluster_acl {
{{- range $node := .Nodes}}
	"{{(index $node.Addresses 0).IP}}";
{{- end}}
}

sub vcl_init {
	new vk8s_cluster = directors.shard();
	{{range $node := .Nodes -}}
	vk8s_cluster.add_backend({{$node.Name}});
	{{end -}}
	vk8s_cluster.reconfigure();
}

sub vcl_recv {
	{{ if .PrimaryOnly -}}
	unset req.http.VK8S-Shard-Primary-Only;
	if (remote.ip !~ vk8s_cluster_acl
	    && "" + vk8s_cluster.backend(resolve=NOW) != server.identity) {
		set req.http.VK8S-Shard-Primary-Only = "true";
		set req.backend_hint = vk8s_cluster.backend(resolve=NOW);
		return (pass);
	}
	else {{ end }}if (remote.ip ~ vk8s_cluster_acl) {
		if (req.http.Host == "vk8s_cluster") {
			if (req.url == "/vk8s_cluster_health") {
				return (synth(200));
			}
			return (synth(404));
		}

		# prevent deadlock for accidental cyclic requests
		set req.hash_ignore_busy = true;

		# if we're async, don't deliver stale
		if (req.http.VK8S-Is-Bgfetch == "true") {
			set req.grace = 0s;
		}

		return (hash);
	}
}
{{ if .PrimaryOnly }}
sub vcl_pass {
	if (req.http.VK8S-Shard-Primary-Only) {
		return (fetch);
	}
}
{{ end }}
sub vcl_backend_fetch {
	{{- if .PrimaryOnly }}
	if (bereq.http.VK8S-Shard-Primary-Only) {
		return (fetch);
	}
	{{- end }}
	if (bereq.retries == 0 
	    && !bereq.uncacheable
	    && remote.ip !~ vk8s_cluster_acl
	    && "" + vk8s_cluster.backend(resolve=NOW) != server.identity) {
		set bereq.backend = vk8s_cluster.backend(resolve=LAZY);
		set bereq.http.VK8S-Is-Bgfetch = bereq.is_bgfetch;
		return (fetch);
	}
}

sub vcl_backend_response {
	{{- if .PrimaryOnly }}
	if (bereq.http.VK8S-Shard-Primary-Only) {
		return (deliver);
	}
	{{- end }}
	if (bereq.backend == vk8s_cluster.backend(resolve=LAZY)) {
	        if (beresp.http.VK8S-Cluster-TTL) {
		        set beresp.ttl = std.duration(
                	    beresp.http.VK8S-Cluster-TTL + "s", 1s);
		        if (beresp.ttl > {{.MaxSecondaryTTL}}) {
			        set beresp.ttl = {{.MaxSecondaryTTL}};
		        }
		        unset beresp.http.VK8S-Cluster-TTL;
	        }
                else {
			set beresp.uncacheable = true;
		}
		return (deliver);
	}
}

sub vcl_backend_error {
	if (bereq.backend == vk8s_cluster.backend(resolve=LAZY)) {
		return (deliver);
	}
}

sub vcl_deliver {
	{{- if .PrimaryOnly }}
	if (req.http.VK8S-Shard-Primary-Only) {
		return (deliver);
	}
	{{- end }}
	unset resp.http.VK8S-Cluster-TTL;
	if (remote.ip ~ vk8s_cluster_acl) {
		if (! obj.uncacheable) {
			set resp.http.VK8S-Cluster-TTL = obj.ttl;
		}
		return (deliver);
	}
}
`

const selfShardName = "self-sharding"

var shardTmpl = template.Must(template.New(selfShardName).
	Parse(selfShardTmplSrc))
